
/*
 * blob.cc
 * Copyright (C) 1999 by John Heidemann
 * $Id: blob.cc,v 1.16 1999/08/15 00:14:52 johnh Exp $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#define _GNU_SOURCE // get snprintf
#include <stdio.h>

#include <stdlib.h>  // atoi
#include <iostream.h>
#include <assert.h>

#include <algorithm> // swap

#include "main.hh"
#include "process_view.hh"
#include "blob.hh"


#if 0
#define BLOB_DEBUG(x) x
#else /* ! 0 */
#define BLOB_DEBUG(x) 
#endif /* 0 */



bool blob::next_right_tending_ = false;
blob::blob_redraw_queue_t blob::redraw_queue_;


blob::blob(process_view *pv) :
	pv_(pv),
	dark_percentage_(0),
	dark_offset_(50),
	H_(0.), S_(0.), B_(0.),
	size_(20),
	x_step_(INITIAL_SLOP),
	y_lows_(NULL), y_highs_(NULL),
	num_(0), 
	alloced_num_(0), 
	sign_(1),
	reached_limit_(0),
	redraw_place_(redraw_queue_.end())
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// start with a small square
	adjust_num(2);
	x_ = 0;
	y_lows_[0] = 0;
	y_highs_[0] = INITIAL_SLOP;
	y_lows_[1] = 0;
	y_highs_[1] = INITIAL_SLOP;
	right_tending_ = next_right_tending_;
	next_right_tending_ = !next_right_tending_;
}

blob::~blob()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (y_lows_)
		delete y_lows_;
	if (y_highs_)
		delete y_highs_;
	if (redraw_place_ != redraw_queue_.end())
		redraw_queue_.erase(redraw_place_);
}

struct flush_drawing_f : public unary_function<blob*,void> {
	void operator()(blob*b) { b->redraw(); };
};

void
blob::flush_drawings()
{
	for_each(redraw_queue_.begin(), redraw_queue_.end(), flush_drawing_f());
}

void
blob::resize_int_array(int **iap, int on, int nn)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (!*iap)
		*iap = new int[nn];
	else if (nn > on) {
		int *nia = new int[nn], *oia = *iap;
		for (int i = 0; i < on; i++)
			nia[i] = oia[i];
		*iap = nia;
		delete oia;
	};
	if (nn > on)
		for (int j = on; j < nn; j++)
			(*iap)[j] = 0;
}

void
blob::adjust_num(int new_num)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// allocate
	if (new_num > alloced_num_) {
		resize_int_array(&y_lows_, num_, new_num);
		resize_int_array(&y_highs_, num_, new_num);
		alloced_num_ = new_num;
	};
	num_ = new_num;
}

void
blob::shift_to_left()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	int i;
	for (i = 0; i < num_-1; i++) {
		y_lows_[i] = y_lows_[i+1];
		y_highs_[i] = y_highs_[i+1];
	};
}

void
blob::shift_to_right()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	int i;
	for (i = num_-1; i > 0; i--) {
		y_lows_[i] = y_lows_[i-1];
		y_highs_[i] = y_highs_[i-1];
	};
}

void
blob::widen_on_left()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	BLOB_DEBUG(cout << "blob::widen_on_left\n");
	x_ -= x_step_;
	adjust_num(num_ + 1);
	shift_to_right();
	// y_lows_[0] = y_highs_[0] = lava_random(y_highs_[0] - y_lows_[0]) + y_lows_[0];   // random
	y_lows_[0] = y_highs_[0] = find_rank_midpoint(0);
}

void
blob::widen_on_right()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	BLOB_DEBUG(cout << "blob::widen_on_right\n");
	adjust_num(num_ + 1);
	int last = num_-1,
		last_last = last - 1;
	assert(last_last >= 0);
	// y_lows_[last] = y_highs_[last] = lava_random(y_highs_[last_last] - y_lows_[last_last]) + y_lows_[last_last]; // random
	y_lows_[last] = y_highs_[last] = find_rank_midpoint(last_last);
}

bool
blob::narrow_on_left()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	if (num_ <= 1)
		return false;
	if (y_lows_[0] < y_highs_[0])
		return false;
	BLOB_DEBUG(cout << "blob::narrow_on_left\n");
	// 0 is unnecessary
	x_ += x_step_;
	shift_to_left();
	adjust_num(num_ - 1);
	return true;
}

bool
blob::narrow_on_right()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	int last = num_-1;
	if (num_ <= 1)
		return false;
	if (y_lows_[last] < y_highs_[last])
		return false;
	BLOB_DEBUG(cout << "blob::narrow_on_right\n");
	// last is unnecessary
	adjust_num(num_ - 1);
	return true;
}

// advance rank returns the amount of delta it was
// NOT able to use.  (0 == success)
int
blob::advance_rank(blob_op op, int num, int delta)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	int *backp, *frontp;
	int abs_delta = delta < 0 ? -delta : delta;
	int unused_delta = 0;

	if (delta == 0)
		return 1;
	if (delta > 0) {
		backp = &y_lows_[num];
		frontp = &y_highs_[num];
	} else {
		frontp = &y_lows_[num];
		backp = &y_highs_[num];
	};
	if (op == SHRINK_OP && y_highs_[num] - y_lows_[num] < abs_delta) {
		// Don't shrink a column into negativeness.
		// Note: we throw away some delta here,
		// so the view doesn't accurately match the real size
		// after this operation.
		int new_abs_delta = y_highs_[num] - y_lows_[num];
		int sign_delta = delta < 0 ? -1 : 1;
		unused_delta = (abs_delta - new_abs_delta) * sign_delta;
		delta = new_abs_delta * sign_delta;
	};

	// move it
	switch (op) {
	case MOVE_OP:
		*frontp += delta;
		*backp += delta;
		break;
	case GROW_OP:
		*frontp += delta;
		size_ += delta;
		break;
	case SHRINK_OP:
		*backp += delta;
		size_ -= delta;
		break;
	};

	// check limits
	int redelta = 0;
	if (delta < 0 && *frontp < 0) {
		redelta = -*frontp;
	} else if (delta > 0 && *frontp > Y_MAX) {
		redelta = - (*frontp - Y_MAX);
	};
	if (redelta) {
		*frontp += redelta;
		*backp += redelta;
		unused_delta = redelta;
		// size is unchanged
	};

	return unused_delta;
}

int
blob::find_rank_midpoint(int num)
{
	if (num < 0)
		num = 0;
	if (num >= num_)
		num = num_ - 1;
	return (y_highs_[num] - y_lows_[num]) / 2 + y_lows_[num];
}

// return the amount two (presumably adjacent) ranks overlap
int
blob::ranks_overlap(int a, int b)
{
	assert(a >= 0 && a < num_);
	assert(b >= 0 && b < num_);
	// first, eliminate no overlap cases
	if (y_highs_[a] < y_lows_[b])
		return y_lows_[b] - y_highs_[a];
	if (y_lows_[a] > y_highs_[b])
		return y_highs_[b] - y_lows_[a];
	// must be overlap
	return 0;
}

void
blob::reverse_direction()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	BLOB_DEBUG(cout << "blob::reverse_direction\n");
	reached_limit_ = 0;
	sign_ *= -1;
}

// partial up returns the amount of unused delta, like advance_rank
// (0==success)
int
blob::partial_op(blob_op op, int delta)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// pick source
	int src = lava_random(num_);

	// figure direction
	// 50% straight, 25% right or left
	int dest_dir = lava_random(4) - 1; if (dest_dir == 2) dest_dir = 0;
	if (src + dest_dir < 0) {
		if (x_ - x_step_ < 0) {  // don't go off edge!
			dest_dir = 0;
		} else {
			widen_on_left();
			src++;
		};
	};
	if (src + dest_dir >= num_) {
		if (x_ + x_step_ * num_ >= X_MAX) {  // watch edge
			dest_dir = 0;
		} else {
			widen_on_right();
		};
	};
	int adj_src = src + dest_dir;

	// second chance---if two adjacent rows are
	// largely askew, move the one that makes things better.
	// Randomly try one side or the other.
	// Initial results suggest that this section of code
	// makes things look a *lot* less stringy than without.
	// (Gee, aparently approximate algorithms based on 
	// local information *are* able to effect globally
	// useful properties... what a good idea :-)
	//
	// TODO: we shouldn't always give things a 2nd chance,
	// but make it porportional to something interesting (like
	// disk IO).
	static int old_first = 1, old_second = -1;
	int first = old_first, second = old_second;
	swap(old_first, old_second);

	int diff;
	if (valid_num(adj_src + first) && (diff = ranks_overlap(adj_src + first, adj_src))) {
		if (sign(diff) == sign(delta))
			adj_src += first;
	} else if (valid_num(adj_src + second) && (diff = ranks_overlap(adj_src + second, adj_src))) {
		if (sign(diff) == sign(delta))
			adj_src += second;
	};

	// advance it
	int unused_delta = advance_rank(op, adj_src, delta);
	if (unused_delta && (op == MOVE_OP || op == GROW_OP)) {
		BLOB_DEBUG(cout << "bounce\n");
		if (++reached_limit_ >= num_ / 2) {
			reverse_direction();
		};
	};

	// check for empty edges
	while (narrow_on_right()) ;
	while (narrow_on_left()) ;

	return unused_delta;
}

void
blob::op(blob_op op, int delta)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	delta *= sign_;

	int delta_sign = sign(delta);
	int abs_delta = delta < 0 ? -delta : delta;

	int i = 0;
	while (abs_delta > 0) {
		int this_delta = min(DELTA_LIMIT, abs_delta);
		int unused_delta = partial_op(op, this_delta * delta_sign);
		abs_delta -= this_delta - unused_delta;
		if (++i > 200) {
			static int uid = -1;
			if (uid == -1)
				uid = getuid();
			if (uid == 2274)  // i.e., me
				cerr << "blob::op: looping limit exceeded, blob " << (void*)this << "\n";
			break;
		};
	};
}

void
blob::print()
{
	ENTRY_TRACE(__FILE__,__LINE__);
	cout << "blob: x=" << x_ << " step=" << x_step_ << " num=" << num_;
	for (int i = 0; i < num_; i++) {
		cout << " " << y_lows_[i] << "-" << y_highs_[i];
	};
	cout << "\n";
}

void
blob::set_hsb(double h, double s, double b)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	H_ = h;
	S_ = s;
	B_ = b;
}

void
blob::set_size(int size)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// fix the size
	this->grow(size - size_);
}

void
blob::set_position(int x, int y)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	// fix the size
	x_ += x;
	for (int i = 0; i < num_; i++) {
		y_lows_[i] += y;
		y_highs_[i] += y;
	};
}

void
blob::info_to_buf(char *buf, int len)
{
	ENTRY_TRACE(__FILE__,__LINE__);
	pv_->info_to_buf(buf, len);
}

blob *
blob::superior()
{
	process_view *superior_view = pv_->superior();
	return superior_view ?  superior_view->get_blob() : NULL;
}

void
blob::set_dark_percentage(int pct, int offset)
{
	dark_percentage_ = pct;
	if (offset != -1) {
		dark_offset_ = offset;
	};
}

